using git_fault;
using LibGit2Sharp;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace TestProject
{
    class UiControllerTester : IRenderer
    {
        public List<(string, string, List<(string, string)>)> faults;
        public string fault;
        public string sha;
        public string appendix;
        public string appendixName;

        public void FaultDeleted(string id, string oldText)
        {
        }

        public void FaultsListed(bool full, List<(string, string, List<(string, string)>)> faults)
        {
            this.faults = faults;
        }

        public void FaultUpdated(string id, string oldText, string newText)
        {
        }

        public void NewFaultCreated(string sha, string fault)
        {
            this.fault = fault;
            this.sha = sha;
        }

        public void ShowAppendix(string appendix, string name)
        {
            this.appendix = appendix;
            this.appendixName = name;
        }

        public void Error(string msg)
        {
            Console.WriteLine(msg);
        }

        public void ShowFault(string fault)
        {
            this.fault = fault;
        }
    }

    [TestClass]
    public class CommandsTest
    {
        string yamlHeader = @"---
Title: a
Priority: 1
Status: Open
---
";

        readonly string path = "./test-repo";
        readonly string pathWd1 = "./test-repo1";
        readonly string pathWd2 = "./test-repo2";
        readonly string barepath = "./test-repo-bare";
        private readonly string byPath = "./test-by-path";

        [TestCleanup]
        public void Cleanup()
        {
            try
            {
                DeleteDirectory(path);
                DeleteDirectory(pathWd1);
                DeleteDirectory(pathWd2);
                DeleteDirectory(barepath);
                DeleteDirectory(byPath);
                Console.WriteLine("Cleanup executed.");
            }
            catch (Exception)
            {
                // ignored
            }
        }

        [TestInitialize]
        public void Setup()
        {
            DeleteDirectory(path);
            DeleteDirectory(pathWd1);
            DeleteDirectory(pathWd2);
            DeleteDirectory(barepath);
            DeleteDirectory(byPath);

            var folder = Directory.CreateDirectory(path);
            var folderWd1 = Directory.CreateDirectory(pathWd1);
            var folderWd2 = Directory.CreateDirectory(pathWd2);
            var folderbare = Directory.CreateDirectory(barepath);
            Directory.CreateDirectory(byPath);

            LibGit2Sharp.Repository.Init(folder.FullName);
            LibGit2Sharp.Repository.Init(folderbare.FullName, true);
            LibGit2Sharp.Repository.Clone(folderbare.FullName, folderWd1.FullName, new CloneOptions());
            LibGit2Sharp.Repository.Clone(folderbare.FullName, folderWd2.FullName, new CloneOptions());

            Console.WriteLine("Setup executed.");
        }

        [TestMethod]
        public void TestAddSingleFaultByText()
        {
            var htmlrenderer = new HtmlRenderer();
            var textrenderer = new TextRenderer();
            var c = new git_fault.CommandHandlers(path, htmlrenderer, textrenderer);

            var teststr = yamlHeader + "This is my first issue";
            teststr = teststr.Replace("\r\n", "\n");

            c.CreateFault(teststr, null, false);
            var repo = new Repository(path);
            var refs = repo.Refs.FromGlob("refs/faults/*");
            Assert.AreEqual(1, refs.Count());
            var text = refs.Single().ResolveToDirectReference().Target.Peel<Commit>().Tree["issue"].Target.Peel<Blob>().GetContentText();
            Assert.AreEqual(teststr, text);
        }

        [TestMethod]
        public void TestAddSingleFaultByPath()
        {
            var htmlrenderer = new HtmlRenderer();
            var textrenderer = new TextRenderer();
            var c = new git_fault.CommandHandlers(path, htmlrenderer, textrenderer);

            var teststr = yamlHeader + "This is my first issue";
            teststr = teststr.Replace("\r\n", "\n");

            File.WriteAllText(Path.Combine(byPath, "issue"), teststr);

            c.CreateFault(null, byPath, false);
            var repo = new Repository(path);
            var refs = repo.Refs.FromGlob("refs/faults/*");
            Assert.AreEqual(1, refs.Count());
            var text = refs.Single().ResolveToDirectReference().Target.Peel<Commit>().Tree["issue"].Target.Peel<Blob>().GetContentText();
            Assert.AreEqual(teststr, text);
        }

        [TestMethod]
        public void TestAddMultipleFaults()
        {
            var htmlrenderer = new HtmlRenderer();
            var textrenderer = new TextRenderer();
            var c = new git_fault.CommandHandlers(path, htmlrenderer, textrenderer);

            var teststr1 = yamlHeader + "This is my first issue";
            teststr1 = teststr1.Replace("\r\n", "\n");
            var teststr2 = yamlHeader + "This is my second issue";
            teststr2 = teststr2.Replace("\r\n", "\n");


            c.CreateFault(teststr1, "", false);
            c.CreateFault(teststr2, "", false);
            var repo = new Repository(path);
            var refs = repo.Refs.FromGlob("refs/faults/*").ToList();
            Assert.AreEqual(2, refs.Count());

            var text1 = refs[0].ResolveToDirectReference().Target.Peel<Commit>().Tree["issue"].Target.Peel<Blob>().GetContentText();
            var text2 = refs[1].ResolveToDirectReference().Target.Peel<Commit>().Tree["issue"].Target.Peel<Blob>().GetContentText();


            bool test = false;
            if (text1 == teststr1 && text2 == teststr2)
            {
                test = true;
            }


            else if (text2 == teststr1 && text1 == teststr2)
            {
                test = true;
            }

            Assert.IsTrue(test);
        }

        [TestMethod]
        public void TestUpdateFault()
        {
            var htmlrenderer = new HtmlRenderer();
            var textrenderer = new TextRenderer();
            var c = new git_fault.CommandHandlers(path, htmlrenderer, textrenderer);

            var teststr1 = yamlHeader + "This is my first issue";
            teststr1 = teststr1.Replace("\r\n", "\n");
            var teststr2 = yamlHeader + "This is my second issue";
            teststr2 = teststr2.Replace("\r\n", "\n");

            c.CreateFault(teststr1, null, false);
            var repo = new Repository(path);
            var refs = repo.Refs.FromGlob("refs/faults/*").ToList();

            var commit1 = refs[0].ResolveToDirectReference().Target.Peel<Commit>();

            c.UpdateFault(commit1.Sha, teststr2, null, false);

            Assert.AreEqual(teststr1, repo.Lookup<Commit>(commit1.Sha).Tree["issue"].Target.Peel<Blob>().GetContentText());
            Assert.AreEqual(teststr2,
                repo.Refs[$"refs/faults/local/{commit1.Sha}"].ResolveToDirectReference().Target.Peel<Commit>().Tree["issue"].Target.Peel<Blob>().GetContentText());
        }

        [TestMethod]
        public void TestDeleteFault()
        {
            var htmlrenderer = new HtmlRenderer();
            var textrenderer = new TextRenderer();
            var c = new git_fault.CommandHandlers(path, htmlrenderer, textrenderer);

            var teststr = yamlHeader + "This is my first issue";
            teststr = teststr.Replace("\r\n", "\n");

            c.CreateFault(teststr, null, false);
            var repo = new Repository(path);
            var refs = repo.Refs.FromGlob("refs/faults/*").ToList();

            var commit1 = refs[0].ResolveToDirectReference().Target.Peel<Commit>();

            c.DeleteFault(commit1.Sha, false);

            Assert.AreEqual(teststr, repo.Lookup<Commit>(commit1.Sha).Tree["issue"].Target.Peel<Blob>().GetContentText());
            Assert.AreEqual(null, repo.Refs[$"refs/faults/local/{commit1.Sha}"].ResolveToDirectReference().Target.Peel<Commit>().Tree["issue"]?.Target.Peel<Blob>().GetContentText());
        }

        [TestMethod]
        public void TestLsFaults()
        {
            var uic = new UiControllerTester();
            var c = new git_fault.CommandHandlers(path, uic, uic);

            c.CreateFault(yamlHeader + "This is my first issue", null, false);
            c.CreateFault(yamlHeader + "This is my second issue", null, false);
            c.CreateFault(yamlHeader + "This is my third issue", null, false);
            var repo = new Repository(path);
            var refs = repo.Refs.FromGlob("refs/faults/*").ToList();

            c.LsFault(false, false);
            Assert.AreEqual(3, uic.faults.Count());
        }

        [TestMethod]
        public void TestShowFault()
        {
            var uic = new UiControllerTester();
            var c = new git_fault.CommandHandlers(path, uic, uic);

            var teststr1 = yamlHeader + "This is my first issue";
            teststr1 = teststr1.Replace("\r\n", "\n");

            c.CreateFault(teststr1, null, false);
            var repo = new Repository(path);
            var refs = repo.Refs.FromGlob("refs/faults/*").ToList();

            c.ShowFault(refs[0].ResolveToDirectReference().Target.Sha, false);
            Assert.AreEqual(teststr1, uic.fault);
        }

        [TestMethod]
        public void TestCreateAppendix()
        {
            var uic = new UiControllerTester();
            var c = new git_fault.CommandHandlers(path, uic, uic);

            c.CreateFault(yamlHeader + "This is my first issue", null, false);

            var name = "testfile.json";
            var repo = new Repository(path);
            var orefA = repo.Refs.FromGlob("refs/faults/*").ToList().First().CanonicalName.Split("/");
            var oref = orefA[orefA.Length - 1];
            c.CreateAppendix(oref, $"../{name}", false);
            var nref = repo.Refs.FromGlob("refs/faults/*").ToList().First();

            var commit = nref.ResolveToDirectReference().Target.Peel<Commit>();
            Assert.IsNotNull(commit.Tree["appendix"]);
            Assert.IsNotNull(commit.Tree["appendix"].Target.Peel<Tree>()[name]);
            var file = commit.Tree["appendix"].Target.Peel<Tree>()[name].Target.Peel<Blob>().GetContentText();
            Assert.AreEqual(File.ReadAllText($"./{name}"), file);
        }

        [TestMethod]
        public void TestShowAppendix()
        {
            var uic = new UiControllerTester();
            var c = new git_fault.CommandHandlers(path, uic, uic);

            c.CreateFault(yamlHeader + "This is my first issue", null, false);

            var name = "testfile.json";
            var repo = new Repository(path);
            var orefA = repo.Refs.FromGlob("refs/faults/*").ToList().First().CanonicalName.Split("/");
            var oref = orefA[orefA.Length - 1];
            c.CreateAppendix(oref, $"../{name}", false);
            var nref = repo.Refs.FromGlob("refs/faults/*").ToList().First();
            var refStrA = nref.CanonicalName.Split("/");
            var refStr = refStrA[refStrA.Length - 1];

            c.ShowAppendix(refStr, "testfile.json", false);

            Assert.AreEqual("testfile.json", uic.appendixName);
            Assert.AreEqual(File.ReadAllText($"./{name}"), uic.appendix);
        }

        [TestMethod]
        public void TestDeleteAppendix()
        {
            var uic = new UiControllerTester();
            var c = new git_fault.CommandHandlers(path, uic, uic);

            c.CreateFault(yamlHeader + "This is my first issue", null, false);

            var name = "testfile.json";
            var repo = new Repository(path);
            var orefA = repo.Refs.FromGlob("refs/faults/*").ToList().First().CanonicalName.Split("/");
            var oref = orefA[orefA.Length - 1];
            c.CreateAppendix(oref, $"../{name}", false);
            var nref = repo.Refs.FromGlob("refs/faults/local/*").ToList().First();


            var refStrA = nref.CanonicalName.Split("/");
            var refStr = refStrA[refStrA.Length - 1];

            c.DeleteAppendix(refStr, "testfile.json", false);
            c.ShowAppendix(refStr, "testfile.json", false);

            Assert.IsNull(uic.appendix);
        }

        [TestMethod]
        public void TestFetch()
        {
            var htmlrenderer = new HtmlRenderer();
            var textrenderer = new TextRenderer();
            var c1 = new git_fault.CommandHandlers(pathWd1, htmlrenderer, textrenderer);
            var c2 = new git_fault.CommandHandlers(pathWd2, htmlrenderer, textrenderer);

            var teststr1 = yamlHeader + "This is my first issue";
            teststr1 = teststr1.Replace("\r\n", "\n");

            c1.CreateFault(teststr1, null, false);
            c1.Push();
            c2.Fetch(false);

            c1.Dispose();
            c2.Dispose();

            using var repo = new Repository(pathWd2);
            var refs = repo.Refs.FromGlob("refs/faults/origin/*");
            Assert.AreEqual(1, refs.Count());
            var text = refs.Single().ResolveToDirectReference().Target.Peel<Commit>().Tree["issue"].Target.Peel<Blob>().GetContentText();
            Assert.AreEqual(teststr1, text);
            repo.Dispose();
        }

        [TestMethod]
        public void TestPushPull()
        {
            var htmlrenderer = new HtmlRenderer();
            var textrenderer = new TextRenderer();
            var c1 = new git_fault.CommandHandlers(pathWd1, htmlrenderer, textrenderer);
            var c2 = new git_fault.CommandHandlers(pathWd2, htmlrenderer, textrenderer);

            var teststr1 = yamlHeader + "This is my first issue";
            teststr1 = teststr1.Replace("\r\n", "\n");

            c1.CreateFault(teststr1, null, false);
            c1.Push();
            c2.Pull(false);

            c1.Dispose();
            c2.Dispose();

            using var repo = new Repository(pathWd2);
            var refs = repo.Refs.FromGlob("refs/faults/local/*");
            Assert.AreEqual(1, refs.Count());
            var text = refs.Single().ResolveToDirectReference().Target.Peel<Commit>().Tree["issue"].Target.Peel<Blob>().GetContentText();
            Assert.AreEqual(teststr1, text);
            repo.Dispose();
        }
        
        [TestMethod]
        public void TestLostUpdate()
        {
            var textrenderer = new UiControllerTester();
            var c1 = new CommandHandlers(pathWd1, textrenderer, textrenderer);
            var c2 = new CommandHandlers(pathWd2, textrenderer, textrenderer);

            var teststr1 = yamlHeader + "This is my first issue";
            teststr1 = teststr1.Replace("\r\n", "\n");

            var teststr2 = yamlHeader + "foobar";
            teststr2 = teststr2.Replace("\r\n", "\n");

            var teststr3 = yamlHeader + "foobar 2";
            teststr3 = teststr3.Replace("\r\n", "\n");

            c1.CreateFault(teststr1, null, false);
            var faultId = textrenderer.sha;

            c1.Push();
            c2.Pull(false);

            c1.UpdateFault(faultId, teststr2, null, false);
            c2.UpdateFault(faultId, teststr3, null, false);

            c1.Push();

            Assert.ThrowsException<NonFastForwardException>(() => c2.Push());
           
            c1.Dispose();
            c2.Dispose();
        }      

        [TestMethod]
        public void TestMerge()
        {
            var textrenderer = new UiControllerTester();
            var c1 = new CommandHandlers(pathWd1, textrenderer, textrenderer);
            var c2 = new CommandHandlers(pathWd2, textrenderer, textrenderer);

            var teststr1 = yamlHeader + "This is my first issue";
            teststr1 = teststr1.Replace("\r\n", "\n");

            var teststr2 = yamlHeader + "foobar";
            teststr2 = teststr2.Replace("\r\n", "\n");

            var teststr3 = yamlHeader + "foobar 2";
            teststr3 = teststr3.Replace("\r\n", "\n");

            c1.CreateFault(teststr1, null, false);
            var faultId = textrenderer.sha;

            c1.Push();

            c2.Pull(false);

            c1.UpdateFault(faultId, teststr2, null, false);
            c2.UpdateFault(faultId, teststr3, null, false);

            c1.Push();
            c2.Pull(false);
            c2.Push();
            
            c1.Pull(false);

            c1.ShowFault(faultId, false);
            var c1FFault = textrenderer.fault;
            
            c2.ShowFault(faultId, false);
            var c2FFault = textrenderer.fault;

            Assert.AreEqual(c1FFault, c2FFault);
            Assert.AreNotEqual(teststr2, teststr3 );
            Assert.AreNotEqual(teststr2, c1FFault );
            
            c1.Dispose();
            c2.Dispose();
        }

        public static void DeleteDirectory(string target_dir)
        {
            if (!Directory.Exists(target_dir))
                return;
            string[] files = Directory.GetFiles(target_dir);
            string[] dirs = Directory.GetDirectories(target_dir);

            foreach (string file in files)
            {
                File.SetAttributes(file, FileAttributes.Normal);
                File.Delete(file);
            }

            foreach (string dir in dirs)
            {
                DeleteDirectory(dir);
            }

            Directory.Delete(target_dir);
        }
    }
}