﻿using System.CommandLine;
using System.CommandLine.Invocation;
using System.CommandLine.Parsing;
using System.IO;

namespace git_fault
{
    internal class ArgParser
    {
        public static RootCommand AddCommandHandlers(CommandHandlers commandHandler)
        {
            var createFaultCommand = new Command("create", "Create fault");
            createFaultCommand.Add(new Option<string>("--text"));
            createFaultCommand.Add(new Option<string>("--path"));
            createFaultCommand.Add(new Option<bool>("--as-html"));

            var updateFaultCommand = new Command("update", "Update fault to list");
            updateFaultCommand.Add(new Option<string>("--id"));
            updateFaultCommand.Add(new Option<string>("--text"));
            updateFaultCommand.Add(new Option<string>("--path"));
            updateFaultCommand.Add(new Option<bool>("--as-html"));

            var deleteFaultCommand = new Command("delete", "Delete fault to list");
            deleteFaultCommand.Add(new Option<string>("--id"));
            deleteFaultCommand.Add(new Option<bool>("--as-html"));

            var lsFaultCommand = new Command("ls", "List faults");
            lsFaultCommand.Add(new Option<bool>("--full"));
            lsFaultCommand.Add(new Option<bool>("--as-html"));

            var showFaultCommand = new Command("show", "Show fault");
            showFaultCommand.Add(new Option<string>("--id"));
            showFaultCommand.Add(new Option<bool>("--as-html"));

            var createAppendixCommand = new Command("create-appendix", "Create an appendix  to the fault. An appendix is a file.");
            var fileOption = new Option<string>("--file");
            fileOption.AddValidator(IsValidFile);
            createAppendixCommand.Add(new Option<string>("--id"));
            createAppendixCommand.Add(fileOption);
            createAppendixCommand.Add(new Option<bool>("--as-html"));

            var deleteAppendixCommand = new Command("delete-appendix", "Delete an appendix from the fault");
            var showAppendixCommand = new Command("show-appendix", "Show an appendix from the fault");
            var fileid = new Option<string>("--faultid");
            fileid.AddValidator((o) => { return ""; });
            showAppendixCommand.Add(fileid);
            showAppendixCommand.Add(new Option<string>("--name"));
            showAppendixCommand.Add(new Option<bool>("--as-html"));

            var pushFaultCommand = new Command("push", "Push fault to remote");
            var pullFaultCommand = new Command("pull", "Pull fault from remote");

            var fetchFaultCommand = new Command("fetch", "Fetch faults from remote");
            var mergeFaultCommand = new Command("merge", "Merge Fault");
            
            var rootCommand = new RootCommand
            {
                new Option<string>(new[] {"--help", "-h"}, "This help."),
                createFaultCommand,
                updateFaultCommand,
                deleteFaultCommand,
                pushFaultCommand,
                pullFaultCommand,
                fetchFaultCommand,
                mergeFaultCommand,
                lsFaultCommand,
                showFaultCommand,
                createAppendixCommand,
                showAppendixCommand,
                deleteAppendixCommand
            };

            createFaultCommand.Handler = CommandHandler.Create<string, string, bool>(commandHandler.CreateFault);
            updateFaultCommand.Handler = CommandHandler.Create<string, string, string, bool>(commandHandler.UpdateFault);
            deleteFaultCommand.Handler = CommandHandler.Create<string, bool>(commandHandler.DeleteFault);
            pushFaultCommand.Handler = CommandHandler.Create(commandHandler.Push);
            fetchFaultCommand.Handler = CommandHandler.Create<bool>(commandHandler.Fetch);
            mergeFaultCommand.Handler = CommandHandler.Create<bool>(commandHandler.Merge);
            pullFaultCommand.Handler = CommandHandler.Create<bool>(commandHandler.Pull);
            lsFaultCommand.Handler = CommandHandler.Create<bool, bool>(commandHandler.LsFault);
            showFaultCommand.Handler = CommandHandler.Create<string, bool>(commandHandler.ShowFault);
            createAppendixCommand.Handler = CommandHandler.Create<string, string, bool>(commandHandler.CreateAppendix);
            showAppendixCommand.Handler = CommandHandler.Create<string, string, bool>(commandHandler.ShowAppendix);
            deleteAppendixCommand.Handler = CommandHandler.Create<string, string, bool>(commandHandler.DeleteAppendix);
            return rootCommand;
        }

        private static string IsValidFile(OptionResult symbolResult)
        {
            var path = symbolResult.Children[0].Tokens[0].Value;
            if (path == null)
            {
                return "No File was specified";
            }

            if (!File.Exists(path))
            {
                return "File does not exists";
            }

            return "";
        }
    }
}