﻿using System.Linq;
using Markdig;
using Markdig.Extensions.Yaml;
using Markdig.Syntax;
using YamlDotNet.Serialization;
using YamlDotNet.Serialization.NamingConventions;

namespace git_fault
{
    public static class Validator
    {
        public static bool AllFieldsInYAMLFilled(string fault)
        {
            var deserializer = new DeserializerBuilder()
                .WithNamingConvention(PascalCaseNamingConvention.Instance)
                .IgnoreUnmatchedProperties()
                .Build();

            MarkdownPipeline Pipeline = new MarkdownPipelineBuilder()
                .UseYamlFrontMatter()
                .Build();

            var document = Markdown.Parse(fault, Pipeline);
            var block = document.Descendants<YamlFrontMatterBlock>().FirstOrDefault();
            if (block == null)
            {
                return false;
            }

            var yaml = block.Lines
                .Lines
                .OrderByDescending(x => x.Line)
                .Select(x => $"{x}\n")
                .ToList()
                .Select(x => x.Replace("---", string.Empty))
                .Select(x => x.Replace("...", string.Empty))
                .Where(x => !string.IsNullOrWhiteSpace(x))
                .Aggregate((s, agg) => agg + s);

            var faultmeta = deserializer.Deserialize<FaultMeta>(yaml);

            if (faultmeta.Priority == null)
            {
                return false;
            }

            if (faultmeta.Status == null)
            {
                return false;
            }

            if (faultmeta.Title == null)
            {
                return false;
            }

            return true;
        }

    }
}